
def dabax_f0_files():
    return ["f0_InterTables.dat", "f0_WaasKirf.dat", "f0_xop.dat"]

def dabax_f1f2_files():
    return ["f1f2_Windt.dat",
            "f1f2_Sasaki.dat",
            "f1f2_Henke.dat",
            "f1f2_EPDL97.dat",
            "f1f2_CromerLiberman.dat",
            "f1f2_Chantler.dat",
            "f1f2_BrennanCowan.dat",
            "f1f2_asf_Kissel.dat",
            ]

def dabax_crosssec_files():
    return ["CrossSec_EPDL97.dat",
            "CrossSec_XCOM.dat",
            "CrossSec_BrennanCowan.dat",
            "CrossSec_StormIsrael.dat",
            "CrossSec_McMaster.dat",
            # "CrossSec_NISTxaamdi.dat",   # they do not contain partial cross sections (compton, rayl, photo)
            # "CrossSec_PE_Scofield.dat",   # they do not contain partial cross sections (compton, rayl, photo)
            ]


"""

dabax: python module for processing remote files containing dabax

"""

import numpy
import os
from urllib.request import urlretrieve
from silx.io.specfile import SpecFile

from dabax.common_tools import calculate_f0_from_f0coeff, f0_interpolate_coefficients
from dabax.common_tools import atomic_symbols, atomic_names, atomic_number
from dabax.common_tools import parse_formula

class DabaxBase(object):
    def __init__(self,
                 dabax_repository=None,
                 verbose=False,
                 file_f0="f0_InterTables.dat",
                 file_f1f2="f1f2_Windt.dat",
                 file_CrossSec = "CrossSec_EPDL97.dat",
                 file_Crystals="Crystals.dat",
                 ):

        self._dabax_repository = dabax_repository
        self._verbose = verbose
        self._file_f0 = file_f0
        self._file_f1f2 = file_f1f2
        self._file_CrossSec = file_CrossSec
        self._file_Crystals = file_Crystals

        if dabax_repository is None:
            self._dabax_repository = self.get_dabax_default_repository()
        else:
            self._dabax_repository = dabax_repository

    def get_dabax_default_repository(self):
        return "https://raw.githubusercontent.com/oasys-kit/DabaxFiles/main/"

    def set_dabax_repository(self, repo):
        self._dabax_repository = repo

    def get_dabax_repository(self):
        return self._dabax_repository

    def set_verbose(self, value=True):
        self._verbose = value

    def verbose(self):
        return self._verbose

    def set_file_f0(self, filename):
        self._file_f0 = filename

    def get_file_f0(self):
        return self._file_f0

    def set_file_f1f2(self, filename):
        self._file_f1f2 = filename

    def get_file_f1f2(self):
        return self._file_f1f2

    def set_file_CrossSec(self, filename):
        self._file_CrossSec = filename

    def get_file_CrossSec(self):
        return self._file_CrossSec

    def set_file_Crystals(self, filename):
        self._file_Crystals = filename

    def get_file_Crystals(self):
        return self._file_Crystals

    def info(self):
        txt = "################  DABAX info ###########\n"
        txt += "dabax repository: %s\n" % self.get_dabax_repository()
        txt += "dabax f0 file: %s\n" % self.get_file_f0()
        txt += "dabax f1f2 file: %s\n" % self.get_file_f1f2()
        txt += "dabax CrossSec file: %s\n" % self.get_file_CrossSec()
        txt += "dabax Crystals file: %s\n" % self.get_file_Crystals()
        txt += "########################################\n"
        return txt


    def is_remote(self):
        if "http" in self.dabax_repository:
            return True
        else:
            return False


    #########################
    # common access tools
    #########################
    def get_dabax_file(self, filename):
        #
        # file exists in current directory
        #
        dabax_repository = self.get_dabax_repository()

        if os.path.exists(filename):
            if self.verbose(): print("Dabax file exists in local directory: %s " % filename)
            return filename
        #
        # download remote file
        #
        if dabax_repository[0:3] == "htt" or dabax_repository[0:3] == "ftp":
            try:
                filepath, http_msg = urlretrieve(dabax_repository + filename,
                                                 filename=filename,
                                                 reporthook=None,
                                                 data=None)

                if self.verbose(): print("Dabax file %s downloaded from %s" % (filepath, dabax_repository + filename))
                return filename
            except:
                try: # in case there are write permissions issues
                    filepath, http_msg = urlretrieve(dabax_repository + filename,
                                                     filename=None, # will create a temp file
                                                     reporthook=None,
                                                     data=None)

                    if self.verbose(): print("Dabax file %s downloaded from %s" % (filepath, dabax_repository + filename))
                    return filepath
                except:
                    raise Exception("Failed to download file %s from %s" % (filename, dabax_repository))
        #
        # file exists in local repository
        #
        f1 = os.path.join(dabax_repository, filename)
        if os.path.exists(f1):
            if self.verbose(): print("Dabax file exists in local directory: %s " % f1)
            return f1

        print("Error trying to access file: %s" % f1)
        raise Exception(FileNotFoundError)


    #########################
    # f0
    #########################

    def get_f0_coeffs_from_dabax_file(self, entry_name="Y3+"):
        filename = self.get_file_f0()
        file1 = self.get_dabax_file(filename)
        if self.verbose(): print("Accessing file: %s" % file1)
        sf = SpecFile(file1)

        flag_found = False

        for index in range(len(sf)):
            s1 = sf[index]
            name = s1.scan_header_dict["S"]

            if name.split('  ')[1] == entry_name:
                flag_found = True
                index_found = index

        if flag_found:
            return (sf[index_found].data)[:, 0]
        else:
            return []


    def f0_with_fractional_charge(self, Z, charge=0.0):
        filename = self.get_file_f0()

        symbol = atomic_symbols()[Z]

        if charge == 0.0:
            return self.get_f0_coeffs_from_dabax_file(entry_name=symbol)
        else:
            # retrieve all entries
            filename = self.get_dabax_file(filename)
            if self.verbose(): print("Accessing file: %s" % filename)
            sf = SpecFile(filename)
            # retrieve all entries
            entries = []
            for index in range(len(sf)):
                s1 = sf[index]
                name = s1.scan_header_dict["S"]
                entries.append(name.split('  ')[1])

            # identify the entries that match the symbol
            interesting_entries = []
            charge_list = []
            index_list = []
            for i, entry in enumerate(entries):
                if entry.find(symbol) == 0:
                    if entry == symbol:
                        interesting_entries.append(entry)
                        charge_list.append(0.0)
                        index_list.append(i)
                    else:
                        entry2 = entry.replace(symbol, '')
                        try:
                            charge_item = int(entry2[::-1])  # convert to integer the reversed string
                            charge_list.append(charge_item)
                            interesting_entries.append(entry)
                            index_list.append(i)
                        except:
                            pass

            # retrieve coefficients from these interesting entries
            coefficient_list = []
            for i in index_list:
                coefficient_list.append((sf[i].data)[:, 0])

            # return self.__f0_interpolate_coefficients(charge, interesting_entries, charge_list, coefficient_list)
            return f0_interpolate_coefficients(charge, charge_list, coefficient_list)

    #
    #
    #

    def _f0_with_fractional_charge_get_entries(self, Z, charge=0.0):
        filename = self.get_file_f0()

        symbol = atomic_symbols()[Z]

        if charge == 0.0:
            return None
        else:
            # retrieve all entries
            filename = self.get_dabax_file(filename)
            if self.verbose(): print("Accessing file: %s" % filename)
            sf = SpecFile(filename)
            # retrieve all entries
            entries = []
            for index in range(len(sf)):
                s1 = sf[index]
                name = s1.scan_header_dict["S"]
                entries.append(name.split('  ')[1])

            # identify the entries that match the symbol
            interesting_entries = []
            charge_list = []
            index_list = []
            for i, entry in enumerate(entries):
                if entry.find(symbol) == 0:
                    if entry == symbol:
                        interesting_entries.append(entry)
                        charge_list.append(0.0)
                        index_list.append(i)
                    else:
                        entry2 = entry.replace(symbol, '')
                        try:
                            charge_item = int(entry2[::-1])  # convert to integer the reversed string
                            charge_list.append(charge_item)
                            interesting_entries.append(entry)
                            index_list.append(i)
                        except:
                            pass

            # retrieve coefficients from these interesting entries
            coefficient_list = []
            for i in index_list:
                coefficient_list.append((sf[i].data)[:, 0])

            # return self.__f0_interpolate_coefficients(charge, interesting_entries, charge_list, coefficient_list)
            return interesting_entries, charge_list, coefficient_list


    ######################
    # f1f2
    ######################

    def f1f2_extract(self, entry_name="Y3+"):
        filename = self.get_file_f1f2()
        file1 = self.get_dabax_file(filename)
        if self.verbose(): print("Accessing file: %s" % filename)
        sf = SpecFile(file1)

        flag_found = False

        for index in range(len(sf)):
            s1 = sf[index]
            name1 = s1.scan_header_dict["S"]
            name = ' '.join(name1.split())
            if name.split(' ')[1] == entry_name:
                flag_found = True
                index_found = index

        if not flag_found:
            if self.verbose():
                print("Entry name %s not found in DABAX file: %s" % (entry_name, filename))
            return None

        # energy
        energy_in_eV = (sf[index_found].data)[0,:].copy()
        if filename == 'f1f2_asf_Kissel.dat' or \
            filename == 'f1f2_Chantler.dat':
            if self.verbose(): print('f1f2_extract: Changing Energy from keV to eV for DABAX file '+filename)
            energy_in_eV *= 1e3

        # f1f2
        if filename == 'f1f2_asf_Kissel.dat':
            f1 = sf[index_found].data[4,:].copy()
            f2 = numpy.abs(sf[index_found].data[1,:].copy())
        else:
            f1 = sf[index_found].data[1,:].copy()
            f2 = sf[index_found].data[2,:].copy()

        return energy_in_eV, f1, f2

    def f1f2_interpolate(self, entry_name, energy,
                         method=2, # 0: lin-lin, 1=lin-log, 2=log-lin, 3:log-log
                         ):

        energy0, f1, f2 = self.f1f2_extract(entry_name)
        if method == 0:
            f1_interpolated = numpy.interp(energy, energy0, f1)
            f2_interpolated = numpy.interp(energy, energy0, f2)
        elif method == 1:
            f1_interpolated = 10 ** numpy.interp((energy),
                                                 (energy0),
                                                 numpy.log10(f1))
            f2_interpolated = 10 ** numpy.interp((energy),
                                                 (energy0),
                                                 numpy.log10(f2))
        elif method == 2:
            f1_interpolated = numpy.interp(numpy.log10(energy),
                                                 numpy.log10(energy0),
                                                 f1)
            f2_interpolated = numpy.interp(numpy.log10(energy),
                                                 numpy.log10(energy0),
                                                 f2)
        elif method == 3:
            f1_interpolated = 10 ** numpy.interp(numpy.log10(energy),
                                                 numpy.log10(energy0),
                                                 numpy.log10(f1))
            f2_interpolated = 10 ** numpy.interp(numpy.log10(energy),
                                                 numpy.log10(energy0),
                                                 numpy.log10(f2))

        return f1_interpolated, f2_interpolated

    ######################
    # crosssec
    ######################

    def crosssec_extract(self, entry_name="Si", partial='TotalCrossSection[barn/atom]'):
        filename = self.get_file_CrossSec()
        file1 = self.get_dabax_file(filename)
        if self.verbose(): print("Accessing file: %s" % filename)
        sf = SpecFile(file1)

        flag_found = False

        for index in range(len(sf)):
            s1 = sf[index]
            name1 = s1.scan_header_dict["S"]
            name = ' '.join(name1.split())
            if name.split(' ')[1] == entry_name:
                flag_found = True
                index_found = index

        if not flag_found:
            if self.verbose():
                print("Entry name %s not found in DABAX file: %s" % (entry_name, filename))
            return None

        data = sf[index_found].data
        labels = sf[index_found].labels

        energy_column_index = -1
        for i in range(len(labels)):
            if 'PhotonEnergy' in labels[i]:
                energy_column_index = i
        if energy_column_index == -1:
            raise Exception("Column with PhotonEnergy not found in scan index %d of %s" % (index_found, file1))

        cs_column_index = -1
        for i in range(len(labels)):
            if partial in labels[i]:
                cs_column_index = i
        if cs_column_index == -1:
            raise Exception("Column with %s not found in scan index %d of %s" % (partial, index_found, file1))

        energy = data[energy_column_index, :].copy()
        cs = data[cs_column_index, :].copy()

        # print("\n>>>>")
        if '[EV]' in labels[energy_column_index].upper():
            pass
        elif '[KEV]' in labels[energy_column_index].upper():
            energy *= 1e3
            # print(">>> energy changed from keV to eV")
        elif '[MEV]' in labels[energy_column_index].upper():
            energy *= 1e6
            # print(">>> energy changed from MeV to eV")

        # print(">>>> %s (col %d)\n     %s (col %d): " % (labels[energy_column_index], energy_column_index, labels[cs_column_index], cs_column_index))

        return energy, cs

    def crosssec_interpolate(self, entry_name, energy,
                         method=2, # 0: lin-lin, 1=lin-log, 2=log-lin, 3:log-log
                         partial='TotalCrossSection[barn/atom]',
                         ):

        out = self.crosssec_extract(entry_name, partial=partial)
        if out is None:
            raise Exception("Descriptor %s not in file %s" % (entry_name, self.get_file_CrossSec()))
        else:
            energy0, cs = out

        if method == 0:
            cs_interpolated = numpy.interp(energy, energy0, cs)
        elif method == 1:
            cs_interpolated = 10 ** numpy.interp((energy),
                                                 (energy0),
                                                 numpy.log10(cs))
        elif method == 2:
            cs_interpolated = numpy.interp(numpy.log10(energy),
                                                 numpy.log10(energy0),
                                                 cs)
        elif method == 3:
            cs_interpolated = 10 ** numpy.interp(numpy.log10(energy),
                                                 numpy.log10(energy0),
                                                 numpy.log10(cs))

        return cs_interpolated

    ######################
    # miscellaneous
    ######################

    def atomic_weights(self, descriptor,
                             filename="AtomicWeights.dat",
                             ):
        """
        ;       Returns atomic weights from DABAX.
        ;
        ; INPUTS:
        ;       id: an identifier string (i.e. 'Si', '70Ge)
        ;
        ;       If descriptor is the symbol (e.g., Ge),
        ;         the averaged atomic mass is returned.
        ;       If descriptor contains the isotope (number of nucleons) (e.g., 70Ge),
        ;         the atomic mass for the isotope is returned.
        ;
        ;       filename = the DABAX  inout file (default AtomicWeights.dat)

        """

        if isinstance(descriptor, str):
            descriptor = [descriptor]
            descriptor_is_string = 1
        else:  # is list
            descriptor_is_string = 0

        # access spec file
        file1 = self.get_dabax_file(filename)
        if self.verbose(): print("Accessing file: %s" % file1)
        sf = SpecFile(file1)

        out = []

        for idescriptor in descriptor:
            flag_found = False
            index_found = []
            for index in range(len(sf)):
                s1 = sf[index]
                name = s1.scan_header_dict["S"]
                line = " ".join(name.split())
                scan_name = line.split(' ')[1]
                if scan_name[-len(idescriptor):] == idescriptor:
                    flag_found = True
                    index_found.append(index)

            if not flag_found:
                raise (Exception("Entry name not found: %s" % idescriptor))

            data = sf[index_found[0]].data

            if idescriptor[0].isdigit():
                out.append(data[0, 0])
            else:
                out.append(data[2, 0])

        if descriptor_is_string:
            return out[0]
        else:
            return out


    def atomic_constants(self, descriptor,
                         filename="AtomicConstants.dat",
                         return_item=0,
                         return_label=None,
                         ):
        """
        ;	Returns atomic constants from DABAX.
        ;
        ; CALLING SEQUENCE:
        ;	out = atomic_constants(id,file,return=return)
        ; INPUTS:
        ;	id: an identifier (or an array of identifiers) to be found in the
        ;	scan title (i.e. 'Si')
        ;
        ; KEYWORDS:
        ;	File = the DABAX  inout file (default: AtomicConstants.dat)
        ;	return_label and return_item  define the variable to be returned.
        ;   If return_name is not None, it has priority over retirn_index
        ;		number of the column in the DABAX file, or a text
        ;		identifier (case insensitive) listed below:
        ;		return_label='AtomicRadius'	             or return_item=0
        ;		return_label='CovalentRadius'	         or return_item=1
        ;		return_label='AtomicMass'	             or return_item=2
        ;		return_label='BoilingPoint'	             or return_item=3
        ;		return_label='MeltingPoint'	             or return_item=4
        ;		return_label='Density'	                 or return_item=5
        ;		return_label='AtomicVolume'	             or return_item=6
        ;		return_label='CoherentScatteringLength'	 or return_item=7
        ;		return_label='IncoherentX-section'	     or return_item=8
        ;		return_label='Absorption@1.8A'	         or return_item=9
        ;		return_label='DebyeTemperature'          or return_item=10
        ;		return_label='ThermalConductivity'       or return_item=11
        ;
        ; OUTPUT:
        ;	out: the value of the selected parameter
        ;
        ; EXAMPLES:
        ;	print(atomic_constants('Si',return='AtomicMass'))
        ;	    28.085500
        ;	print(atomic_constants(14,return='AtomicMass'))
        ;           28.085500
        ;	print(atomic_constants([14,27],return='AtomicMass'))
        ;	    28.085500       58.933200
        ;
        ;-

        """

        if isinstance(descriptor, str):
            descriptor = [descriptor]
            descriptor_is_string = 1
        else:  # is list
            descriptor_is_string = 0

        return_index = -1
        if return_label is None:
            return_index = return_item
        else:
            if return_label == 'AtomicRadius'	: return_index = 0
            if return_label == 'CovalentRadius'	            : return_index = 1
            if return_label == 'AtomicMass'	                : return_index = 2
            if return_label == 'BoilingPoint'	            : return_index = 3
            if return_label == 'MeltingPoint'	            : return_index = 4
            if return_label == 'Density'	                : return_index = 5
            if return_label == 'AtomicVolume'	            : return_index = 6
            if return_label == 'CoherentScatteringLength'	: return_index = 7
            if return_label == 'IncoherentX-section'	    : return_index = 8
            if return_label == 'Absorption@1.8A'	        : return_index = 9
            if return_label == 'DebyeTemperature'           : return_index = 10
            if return_label == 'ThermalConductivity'        : return_index = 11


        if return_index == -1: raise Exception("Bad item index")
        # access spec file

        file1 = self.get_dabax_file(filename)
        if self.verbose(): print("Accessing file: %s" % file1)
        sf = SpecFile(file1)

        out = []

        for idescriptor in descriptor:

            for index in range(len(sf)):
                flag_found = False
                s1 = sf[index]
                name = s1.scan_header_dict["S"]
                scan_name = name.split('  ')[1]
                if scan_name == idescriptor:
                    flag_found = True
                    index_found = index
                    break

            if flag_found:
                out.append((sf[index_found].data)[return_index, 0])
            else:
                raise Exception("Data not found for %s " % idescriptor)

        if descriptor_is_string:
            return out[0]
        else:
            return out


    def element_density(self, descriptor,
                        filename="AtomicConstants.dat"):

        return self.atomic_constants(descriptor, filename=filename, return_label="Density")


    def compound_parser(self, descriptor):

        zetas, fatomic = parse_formula(formula=descriptor, verbose=self.verbose())

        elements = []
        atomic_weight = []
        massFractions = []

        for i ,z in enumerate(zetas):
            symbol = atomic_symbols()[z]
            atw = self.atomic_weights(symbol)
            elements.append(z)
            atomic_weight.append(atw)
            massFractions.append(fatomic[i ] *atw)

        mweight = 0.0
        for i in range(len(fatomic)):
            mweight += atomic_weight[i] * fatomic[i]

        for i in range(len(massFractions)):
            massFractions[i] /= mweight

        new_dict = {
            "nElements": len(elements),
            "nAtomsAll": float(numpy.array(fatomic).sum()),
            "Elements" :zetas,
            "massFractions": massFractions,
            "nAtoms" :fatomic,
            "molarMass": mweight,
        }

        return new_dict

if __name__ == '__main__':
    dx = DabaxBase(dabax_repository="https://gitlab.esrf.fr/srio/dabaxfiles/-/raw/main/")
    print(dx.info())
    #
    # f0
    #
    if False:
        #
        # test f0 data for B3+
        #
        q = numpy.array([0,0.05,0.1,0.15,0.2,0.25,0.3,0.35,0.4,0.5,0.6,0.7,0.8,0.9,1,1.1,1.2,1.3,1.4,1.5,1.6,1.7,1.8,1.9])
        f0_B3plus = numpy.array([2,1.995,1.979,1.954,1.919,1.875,1.824,1.766,1.703,1.566,1.42,1.274,1.132,0.999,0.877,0.767,0.669,0.582,0.507,0.441,0.384,0.335,0.293,0.256])

        #
        # plot
        #
        from srxraylib.plot.gol import plot

        plot(q, f0_B3plus,
             q, calculate_f0_from_f0coeff(dx.f0_with_fractional_charge(5, 3.0), q),
             q, calculate_f0_from_f0coeff(dx.f0_with_fractional_charge(5, 2.8), q),
             xtitle=r"q (sin $\theta$ / $\lambda$)", ytitle="f0 [electron units]",
             legend=["B3plus original",
                     "B3plus from f0_with_fractional_charge(5,+3)",
                     "B3plus from f0_with_fractional_charge(5,+2.8)"],
             marker=['+', None, None],
             title="")

    #
    # f0 another test
    #
    if False:
        #
        # test f0 data for B3+
        #
        q = numpy.array(
            [0, 0.05, 0.1, 0.15, 0.2, 0.25, 0.3, 0.35, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1, 1.1, 1.2, 1.3, 1.4, 1.5, 1.6,
             1.7, 1.8, 1.9])
        f0_B3plus = numpy.array(
            [2, 1.995, 1.979, 1.954, 1.919, 1.875, 1.824, 1.766, 1.703, 1.566, 1.42, 1.274, 1.132, 0.999, 0.877, 0.767,
             0.669, 0.582, 0.507, 0.441, 0.384, 0.335, 0.293, 0.256])

        #
        # plot
        #
        from srxraylib.plot.gol import plot

        plot(q, f0_B3plus,
             q, calculate_f0_from_f0coeff(dx.f0_with_fractional_charge(5, 3.0), q),
             q, calculate_f0_from_f0coeff(dx.f0_with_fractional_charge(5, 2.8), q),
             xtitle=r"q (sin $\theta$ / $\lambda$)", ytitle="f0 [electron units]",
             legend=["B3plus original",
                     "B3plus from f0_with_fractional_charge(5,+3)",
                     "B3plus from f0_with_fractional_charge(5,+2.8)"],
             marker=['+',None,None],
             title="", show=1)


    if False:
        #
        # misc
        #
        print("Ge, Si: ", dx.atomic_weights(["Ge","Si"]))
        print("70Ge: ", dx.atomic_weights("70Ge"))

        print(atomic_symbols()[14], atomic_names()[14])

        print("Si atomic mass", dx.atomic_constants("Si", return_item=2))
        print("Si,Ge atomic mass", dx.atomic_constants(["Si", "Ge"], return_item=2))
        print("Si,Co atomic mass", dx.atomic_constants(["Si", "Co"], return_label='AtomicMass'))

        print("Z=27", atomic_symbols()[27])
        print("Ge Z=%d" % atomic_number("Ge"))

        print("Density Si: ", dx.element_density("Si"))

    if False:
        energy, f1, f2 = dx.f1f2_extract("Si")

        energy_i = numpy.linspace(10,15000,200)
        f1_i, f2_i = dx.f1f2_interpolate("Si", energy=energy_i)
        print(">>>>", energy.shape, f1.shape, f2.shape)
        from srxraylib.plot.gol import plot
        plot(energy, f1,
             energy, f2,
             energy_i, f1_i,
             energy_i, f2_i,
             xlog=True, ylog=True, title="f1f2 Si",
             legend=['f1','f2','f1_i','f2_i'],
             marker=[None,None,'+','+'],
             linestyle=[None,None,'',''])


    if False: # used to create f0_xop_with_fractional_charge_data() in common_tools
        filename = dx.get_file_f0()
        file1 = dx.get_dabax_file(filename)
        if self.verbose(): print("Accessing file: %s" % filename)
        sf = SpecFile(file1)
        for Z in range(1,99):
            interesting_entries, charge_list, coefficient_list = dx._f0_with_fractional_charge_get_entries(Z, charge=1)

            for i,iel in enumerate(coefficient_list):
                coefficient_list[i] = iel.tolist()

            print("    a.append({'Z':",Z,",'charge_list':",charge_list,",'coefficient_list':",coefficient_list,"})" )

    if False:
        from common_tools import f0_xop_with_fractional_charge
        print("f0 coeffs for Z=14 charge=1.5 DABAX/common_tools: ",
              dx.f0_with_fractional_charge(14, charge=1.5),
              f0_xop_with_fractional_charge(14, charge=1.5),)


    if True:
        energy, cs = dx.crosssec_extract("Si")

        energy_i = numpy.linspace(10,15000,200)
        cs_i = dx.crosssec_interpolate("Si", energy=energy_i)
        print(">>>>", energy.shape, cs.shape)
        from srxraylib.plot.gol import plot
        plot(energy, cs,
             energy_i, cs_i,
             xlog=True, ylog=True, title="crosssec Si",
             legend=['cs','cs_i'],
             marker=[None,'+'],
             linestyle=[None,''])


        print(">>>> cs of Si at 10 kev %g barn/atom ", dx.crosssec_interpolate("Si", energy=50000.0))


import numpy
from dabax.dabax_base import DabaxBase
from dabax.dabax_xraylib_decorator import DabaxXraylibDecorator

class DabaxXraylib(DabaxBase, DabaxXraylibDecorator):
    """
    Dabax decorated with xraylib interface (warning: not all xraylib functions are implemented)
    """
    def __init__(self,
                 dabax_repository=None,
                 file_f0="f0_InterTables.dat",
                 file_f1f2="f1f2_Windt.dat",
                 file_CrossSec = "CrossSec_EPDL97.dat",
                 file_Crystals="Crystals.dat",
                 ):

        DabaxBase.__init__(self,
                           dabax_repository=dabax_repository,
                           file_f0=file_f0,
                           file_f1f2=file_f1f2,
                           file_CrossSec=file_CrossSec,
                           file_Crystals=file_Crystals)

if __name__ == "__main__":

    import xraylib
    dx = DabaxXraylib()

    if False:
        #
        # dabax vs xraylib tests
        #

        #
        # crystal tests
        #

        print("DABAX crystal list: \n",        dx.Crystal_GetCrystalsList())
        print("XRAYLIB crystal list: \n", xraylib.Crystal_GetCrystalsList())

        siD =      dx.Crystal_GetCrystal('Si')
        siX = xraylib.Crystal_GetCrystal('Si')

        print("DABAX crystal si: \n",        dx.Crystal_GetCrystal('Si'))
        print("XRAYLIB crystal si: \n", xraylib.Crystal_GetCrystal('Si'))

        print("Si 111 d-spacing: DABAX: %g, XRAYLIB: %g "% \
              (dx.Crystal_dSpacing(siD,1,1,1),xraylib.Crystal_dSpacing(siX,1,1,1)))

        print("Si 111 bragg angle at 10 keV [deg]: DABAX: %g, XRAYLIB: %g "% (\
              180 / numpy.pi * dx.Bragg_angle(siD,10, 1,1,1), \
              180 / numpy.pi * xraylib.Bragg_angle(siX, 10, 1, 1, 1)))

        # F0 = dx.Crystal_F_H_StructureFactor(siD,8.0,0,0,0,1.0,ratio_theta_thetaB=1.0)
        dabax_all_F = dx.Crystal_F_0_F_H_F_H_bar_StructureFactor(siD,8.0,1,1,1,1.0,rel_angle=1.0)

        print("F0 dabax, xraylib: ",
              dx.Crystal_F_H_StructureFactor(siD,8.0,0,0,0,1.0,1.0), dabax_all_F[0],
              xraylib.Crystal_F_H_StructureFactor(siX,8.0,0,0,0,1.0,1.0))

        print("F111 dabax, xraylib: ",
              dx.Crystal_F_H_StructureFactor     (siD,8.1,1,1,1,1.0,1.0), dabax_all_F[1],
              xraylib.Crystal_F_H_StructureFactor(siX,8.1,1,1,1,1.0,1.0))

        print("F-1-1-1 dabax, xraylib: ",
              dx.Crystal_F_H_StructureFactor     (siD,8.1,-1,-1,-1,1.0,1.0), dabax_all_F[2],
              xraylib.Crystal_F_H_StructureFactor(siX,8.1,-1,-1,-1,1.0,1.0))

        #
        # basic tools
        #
        # TODO: does not work for double parenthesis "Ga2(F(KI))3"
        for descriptor in ["H2O","Eu2H2.1O1.3","PO4", "Ca5(PO4)3.1F"]:
            print("\ncompound parsing for %s" % descriptor)
            print("DABAX: ",        dx.CompoundParser(descriptor))
            print("XRAYLIB: ", xraylib.CompoundParser(descriptor))

        print("Si is Z= %d (DABAX)  %d (XRAYLIB)" % (dx.SymbolToAtomicNumber("Si"),xraylib.SymbolToAtomicNumber("Si")))
        print("Z=23 is %s (DABAX)  %s (XRAYLIB)" % (dx.AtomicNumberToSymbol(23),xraylib.AtomicNumberToSymbol(23)))
        print("Density Z=30 %g (DABAX)  %g (XRAYLIB)" % (dx.ElementDensity(30),xraylib.ElementDensity(30)))
        print("AtWeight Z=30 %g (DABAX)  %g (XRAYLIB)" % (dx.AtomicWeight(30),xraylib.AtomicWeight(30)))


        #
        # NIST compounds
        #

        nist_x = xraylib.GetCompoundDataNISTList()
        nist_d = dx.GetCompoundDataNISTList()

        for i in range(len(nist_x)):
            # print("\n")
            # print("\n**%s**  **%s**" % (nist_x[i],nist_d[i]))
            # print (type(nist_x[i]) , type(nist_d[i]))
            if not ( nist_x[i]  in nist_d[i]):
                print(">>>> Error processing %s" % (nist_x[i]))
            print (dx.GetCompoundDataNISTByName(nist_x[i]))


        for i in range(len(nist_x)):
            print(nist_d[i], nist_x[i])
            a_d =      dx.GetCompoundDataNISTByIndex(i)
            a_x = xraylib.GetCompoundDataNISTByIndex(i)
            print("\n\n\n", i, "\n", a_d, "\n", a_x,)
            name_x = a_x["name"]
            name_d = a_d["name"]

            print(name_d, name_x)
            b_d =      dx.GetCompoundDataNISTByName(name_d)
            b_x = xraylib.GetCompoundDataNISTByName(name_x)
            print("\n", b_d, "\n", b_x,)


        print(dx.CompoundParserCheckingNIST("H2O"))
        print(dx.CompoundParserCheckingNIST("Water, Liquid"))

        #
        # scattering factors
        #

        print("Fi  dabax,xraylib: ",  dx.Fi (14,18.0), xraylib.Fi (14,18.0))
        print("Fii dabax,xraylib: ", dx.Fii(14,18.0), xraylib.Fii(14,18.0))

        print("FF_rayl dabax, xraylib: ",
              dx.FF_Rayl(17, 2.2),xraylib.FF_Rayl(17, 2.2) )
        print("FF_Rayl  dabax,xraylib: ", dx.FF_Rayl(14, 2.0), xraylib.FF_Rayl(14, 2.0))

        # loops
        energies = numpy.linspace(15,18,10)
        f1f2_d = numpy.array(dx.FiAndFii(14,energies))
        print(f1f2_d.shape)
        for i,energy in enumerate(energies):
            print("energy = %g" %energy)
            print("   Fi  dabax,xraylib: ",  f1f2_d[0,i], xraylib.Fi (14,energy))
            print("   Fii dabax,xraylib: ",  f1f2_d[1,i], xraylib.Fii(14,energy))


        from dabax.dabax_files import dabax_f1f2_files
        for file in dabax_f1f2_files():
            dx1 = DabaxXraylib(file_f1f2=file)
            print("\nFi  dabax (%s): %g ,xraylib: %g" % (file, dx1.Fi(14, 18.0), xraylib.Fi(14, 18.0)))
            print("Fii  dabax (%s): %g ,xraylib: %g" % (file, dx1.Fii(14, 18.0), xraylib.Fii(14, 18.0)))


        #
        # cross sections
        #

        from dabax.dabax_files import dabax_crosssec_files

        for file in dabax_crosssec_files():
            dx1 = DabaxXraylib(file_CrossSec=file)
            print(">>>>>>>>>>>>>>>  file: ", file)
            try:
                print("CSb_Total Si dabax,xraylib: ",  dx1.CSb_Total(14, 18.0), xraylib.CSb_Total(14, 18.0))
                print("CSb_Photo Si dabax,xraylib: ",  dx1.CSb_Photo(14,18.0),  xraylib.CSb_Photo(14,18.0))
                print("CSb_Rayl  Si dabax,xraylib: ",  dx1.CSb_Rayl (14,18.0),  xraylib.CSb_Rayl (14,18.0))
                print("CSb_Compt Si dabax,xraylib: ",  dx1.CSb_Compt(14,18.0),  xraylib.CSb_Compt(14,18.0))

                print("CS_Total Si dabax,xraylib: ",  dx1.CS_Total(14, 18.0), xraylib.CS_Total(14, 18.0))
                print("CS_Photo Si dabax,xraylib: ",  dx1.CS_Photo(14,18.0),  xraylib.CS_Photo(14,18.0))
                print("CS_Rayl  Si dabax,xraylib: ",  dx1.CS_Rayl (14,18.0),  xraylib.CS_Rayl (14,18.0))
                print("CS_Compt Si dabax,xraylib: ",  dx1.CS_Compt(14,18.0),  xraylib.CS_Compt(14,18.0))
            except:
                print("!!!!!!!!!!Errors with file", file)


        for file in dabax_crosssec_files():
            dx1 = DabaxXraylib(file_CrossSec=file)
            print(">>>>>>>>>>>>>>>  file: ", file)

            try:
                print("CSb_Total SiO2 dabax,xraylib: ",  dx1.CSb_Total_CP("SiO2", 18.0), xraylib.CSb_Total_CP("SiO2", 18.0))
                print("CSb_Photo SiO2 dabax,xraylib: ",  dx1.CSb_Photo_CP("SiO2",18.0),  xraylib.CSb_Photo_CP("SiO2",18.0))
                print("CSb_Rayl  SiO2 dabax,xraylib: ",   dx1.CSb_Rayl_CP("SiO2",18.0),   xraylib.CSb_Rayl_CP("SiO2",18.0))
                print("CSb_Compt SiO2 dabax,xraylib: ",  dx1.CSb_Compt_CP("SiO2",18.0),  xraylib.CSb_Compt_CP("SiO2",18.0))

                print("CS_Total SiO2 dabax,xraylib: ",  dx1.CS_Total_CP("SiO2", 18.0), xraylib.CS_Total_CP("SiO2", 18.0))
                print("CS_Photo SiO2 dabax,xraylib: ",  dx1.CS_Photo_CP("SiO2",18.0),  xraylib.CS_Photo_CP("SiO2",18.0))
                print("CS_Rayl  SiO2 dabax,xraylib: ",   dx1.CS_Rayl_CP("SiO2",18.0),   xraylib.CS_Rayl_CP("SiO2",18.0))
                print("CS_Compt SiO2 dabax,xraylib: ",  dx1.CS_Compt_CP("SiO2",18.0),  xraylib.CS_Compt_CP("SiO2",18.0))
            except:
                print("!!!!!!!!!!Errors with file", file)


        #
        # refractive index
        #
        dens = dx.element_density("Be")
        print("Refractive_Index_Re Be  dabax,xraylib: ",  dx.Refractive_Index_Re ("Be",18.0, dens), xraylib.Refractive_Index_Re ("Be",18.0, dens))
        print("Refractive_Index_Im Be  dabax,xraylib: ",  dx.Refractive_Index_Im ("Be",18.0, dens), xraylib.Refractive_Index_Im ("Be",18.0, dens))

        # loops
        energies = numpy.linspace(15,18,10)
        r1 = dx.Refractive_Index_Re("Be", energies, dens)
        for i,energy in enumerate(energies):
            print("   delta @ %g keV, dabax: %g ,xraylib: %g" % (energy, 1-r1[i], 1-xraylib.Refractive_Index_Re ("Be",energy, dens)))

    else:
        pass





#
# dabax functions with the same interface as xraylib
#
import numpy
import scipy.constants as codata
from silx.io.specfile import SpecFile
from dabax.common_tools import atomic_symbols, atomic_names, atomic_number
from dabax.common_tools import bragg_metrictensor
from dabax.common_tools import calculate_f0_from_f0coeff

class DabaxXraylibDecorator(object):

    #########################
    # crystals
    #########################
    def Crystal_GetCrystal(self, entry_name='YB66'):
        """
        parse a complex crystal structure file into a dictionary (like xraylib.Crystal_GetCrystal('Si'))

        it has an additional fiels for each atom: the charge

        return a dictionary containing crystal infomation
        """
        dabax_repository = self.get_dabax_repository()

        filename = self.get_file_Crystals()

        file1 = self.get_dabax_file(filename)

        if self.verbose(): print("Accessing file: %s" % file1)
        sf = SpecFile(file1)

        flag_found = False

        for index in range(len(sf)):
            s1 = sf[index]
            name = s1.scan_header_dict["S"]
            if name.split(' ')[1] == entry_name:
                flag_found = True
                index_found = index

        if not flag_found:
            raise (Exception("Entry name not found: %s" % entry_name))

        cryst = {'name': entry_name}  # returned dictionary like that one created by xraylib.Crystal_GetCrystal(descriptor)

        cell_parameters = sf[index_found].scan_header_dict["UCELL"]
        cell_parameters = ' '.join(cell_parameters.split())  # remove multiple blanks

        a = cell_parameters.split(' ')
        cryst['a'] = float(a[0])
        cryst['b'] = float(a[1])
        cryst['c'] = float(a[2])
        cryst['alpha'] = float(a[3])
        cryst['beta'] = float(a[4])
        cryst['gamma'] = float(a[5])

        volume = bragg_metrictensor(float(a[0]), float(a[1]), float(a[2]),
                                    float(a[3]), float(a[4]), float(a[5]), RETURN_VOLUME=1)

        cryst['volume'] = volume

        cell_data = numpy.array(sf[index_found].data)

        cryst['n_atom'] = cell_data.shape[1]
        atom = []

        for i in range(cell_data.shape[1]):
            if cell_data.shape[0] == 5:  # standard 5 columns
                # not here, this info is not in the dabax file
                # s = symbol_to_from_atomic_number(int(cell_data[0,i]))
                atom.append({
                    'Zatom': int(cell_data[0, i]),
                    'fraction': cell_data[1, i],
                    'x': cell_data[2, i],
                    'y': cell_data[3, i],
                    'z': cell_data[4, i],
                    'charge': 0.0, })
            else:  # 6 columns (charge)
                # 'AtomicName' required to compatible my current code
                # s = symbol_to_from_atomic_number(int(cell_data[0,i]))
                # if cell_data[5, i] != 0:  #charged
                #     s = s + f'%+.6g'%cell_data[5, i]
                atom.append({
                    # 'AtomicName': s,
                    'Zatom': int(cell_data[0, i]),
                    'fraction': cell_data[1, i],
                    'x': cell_data[2, i],
                    'y': cell_data[3, i],
                    'z': cell_data[4, i],
                    'charge': cell_data[5, i], })

        cryst['atom'] = atom
        cryst['cpointer'] = None

        ANISO_KEY = "UANISO_COFF"  # prefix for a line with anisotropic coefficients
        d = sf[index_found].scan_header_dict
        AnisoItem = {'Name': '       ',
                     'start': 0,
                     'end': 0,
                     'beta11': 0.0,
                     'beta22': 0.0,
                     'beta33': 0.0,
                     'beta12': 0.0,
                     'beta13': 0.0,
                     'beta23': 0.0}

        a = [(x, d[x].split()) for x in d if x[:len(ANISO_KEY)] == ANISO_KEY]
        if len(a) > 0:  # found Anisotropic coefficients in the header, process it
            a = sorted(a, key=lambda x: int(x[1][0]),
                       reverse=False)  # sort 'Start' ascendant, avoid order changed by the SpecFile
            n = 0
            Aniso = []
            for x in a:  # tuple('UANISO_COFF_B1',[1 96 0.00038 0.00044 0.00039 0 0 0])
                AnisoItem['Name'] = x[0][len(ANISO_KEY) + 1:]  # get site atom name starting from 13th character 'B1', etc
                AnisoItem['start'] = int(x[1][0])
                AnisoItem['end'] = int(x[1][1])
                AnisoItem['beta11'] = float(x[1][2])
                AnisoItem['beta22'] = float(x[1][3])
                AnisoItem['beta33'] = float(x[1][4])
                AnisoItem['beta12'] = float(x[1][5])
                AnisoItem['beta13'] = float(x[1][6])
                AnisoItem['beta23'] = float(x[1][7])
                Aniso.append(AnisoItem.copy())
                n = n + 1
            cryst['Aniso'] = Aniso  # if having key 'Ansio' when there is anisotropic data,otherwise no
            cryst['n_aniso'] = n
        else:  # create a dummy Aniso to compatible with xraylib
            cryst['Aniso'] = [AnisoItem.copy()]
            cryst['n_aniso'] = 1

        return cryst


    def Crystal_GetCrystalsList(self):
        """
        get crystal names from crystals.dat
        """
        file1 = self.get_dabax_file(self.get_file_Crystals())
        if self.verbose(): print("Accessing file: %s" % file1)
        sf = SpecFile(file1)
        crystals = []
        for index in range(len(sf)):
            s1 = sf[index]
            name = s1.scan_header_dict["S"]
            crystals.append(name.split(' ')[1])

        return crystals

    def Crystal_dSpacing(self, cryst, h, k, l):
        return bragg_metrictensor(cryst['a'], cryst['b'], cryst['c'],
                                  cryst['alpha'], cryst['beta'], cryst['gamma'],
                                  HKL=[h, k, l])

    def Bragg_angle(self, cryst, E_keV, h, k, l):
        dspacing = self.Crystal_dSpacing(cryst, h, k, l)  # in A
        wavelength = codata.h * codata.c / codata.e / (E_keV * 1e3) * 1e10  # in A
        return numpy.arcsin(wavelength / 2 / dspacing)

    def Crystal_F_H_StructureFactor(self,
                                    crystal_id,
                                    energy_in_kev,
                                    millerH,
                                    millerK,
                                    millerL,
                                    debyeWaller,
                                    ratio_theta_thetaB=1.0):
        energy = energy_in_kev * 1e3
        wavelength = codata.h * codata.c / codata.e / energy * 1e10
        # print(crystal_id["n_atom"])
        atom = crystal_id['atom']
        natom = len(atom)
        list_fraction = [atom[i]['fraction'] for i in range(natom)]
        list_x = [atom[i]['x'] for i in range(natom)]
        list_y = [atom[i]['y'] for i in range(natom)]
        list_z = [atom[i]['z'] for i in range(natom)]

        F_H = numpy.zeros(numpy.array(energy).size, dtype=complex)

        for i in range(natom):
            atom_i = atom[i]
            if (i > 0) and (atom_i['Zatom'] == Z_i) and (atom_i['charge'] == charge_i):
                pass # avoid re-calculating f0 if inputs are identical to previous call
            else:
                Z_i = atom_i['Zatom']
                charge_i = atom_i['charge']
                coeffs = self.f0_with_fractional_charge(Z_i, charge=charge_i)
                if (millerH == 0 and millerK == 0 and millerL == 0):
                    ratio = 0.0
                else:
                    angle = self.Bragg_angle(crystal_id, energy_in_kev,
                                             millerH, millerK, millerL) * ratio_theta_thetaB
                    ratio = numpy.sin(angle) / wavelength
                f0_i = calculate_f0_from_f0coeff(coeffs, ratio)
                Fi  = self.Fi(Z_i, energy_in_kev)
                Fii = self.Fii(Z_i, energy_in_kev)

            F_H += (f0_i + Fi - Fii * 1j) * list_fraction[i] * \
                   numpy.exp(2j*numpy.pi*(millerH*list_x[i]+millerK*list_y[i]+millerL*list_z[i]))

        return F_H * debyeWaller

    #########################
    #  misc
    #########################
    def CompoundParser(self, descriptor):
        return self.compound_parser(descriptor)

    def SymbolToAtomicNumber(self, symbol):
        return atomic_number(symbol)

    def AtomicNumberToSymbol(self, Z):
        return atomic_symbols()[Z]

    def ElementDensity(self, Z):
        return self.element_density(self.AtomicNumberToSymbol(Z))

    def AtomicWeight(self, Z):
        return self.atomic_weights(self.AtomicNumberToSymbol(Z))

    #########################
    #  scattering functions
    #########################
    def Fi(self, Z, energy):
        return self.FiAndFii(Z, energy)[0]

    def Fii(self, Z, energy):
        return self.FiAndFii(Z, energy)[1]

    def FF_Rayl(self, Z, q):

        coeffs = self.f0_with_fractional_charge(Z, charge=0.0)
        return calculate_f0_from_f0coeff(coeffs, q)

    #########################
    #  cross sections
    #########################

    # main (barns)
    def CSb_Total(self, Z, energy):
        return self.crosssec_interpolate(self.AtomicNumberToSymbol(Z), energy * 1e3,
                                         partial='TotalCrossSection[barn/atom]',)

    def CSb_Photo(self, Z, energy):
        return self.crosssec_interpolate(self.AtomicNumberToSymbol(Z), energy * 1e3,
                                         partial='PhotoElectric[barn/atom]',)
    def CSb_Rayl(self, Z, energy):
        return self.crosssec_interpolate(self.AtomicNumberToSymbol(Z), energy * 1e3,
                                         partial='Rayleigh(coherent)[barn/atom]',)

    def CSb_Compt(self, Z, energy):
        return self.crosssec_interpolate(self.AtomicNumberToSymbol(Z), energy * 1e3,
                                         partial='Compton(incoherent)[barn/atom]',)

    # in cm2/g

    def CS_Total(self, Z, energy):
        return self.CSb_Total(Z, energy) * (codata.Avogadro * 1e-24 / self.AtomicWeight(Z))

    def CS_Photo(self, Z, energy):
        return self.CSb_Photo(Z, energy) * (codata.Avogadro * 1e-24 / self.AtomicWeight(Z))

    def CS_Rayl(self, Z, energy):
        return self.CSb_Rayl(Z, energy) * (codata.Avogadro * 1e-24 / self.AtomicWeight(Z))

    def CS_Compt(self, Z, energy):
        return self.CSb_Compt(Z, energy) * (codata.Avogadro * 1e-24 / self.AtomicWeight(Z))


    # for compounds


    def CS_Total_CP(self, descriptor, energy):
        cp = self.CompoundParserCheckingNIST(descriptor,)
        out = 0.0
        for i in range(cp["nElements"]):
            out += self.CS_Total(cp["Elements"][i], energy) * cp["massFractions"][i]
        return out

    def CSb_Total_CP(self, descriptor, energy):
        cp = self.CompoundParserCheckingNIST(descriptor,)
        out = 0.0
        for i in range(cp["nElements"]):
            out += self.CSb_Total(cp["Elements"][i], energy) * cp["massFractions"][i]
        return out

    def CS_Photo_CP(self, descriptor, energy):
        cp = self.CompoundParserCheckingNIST(descriptor,)
        out = 0.0
        for i in range(cp["nElements"]):
            out += self.CS_Photo(cp["Elements"][i], energy) * cp["massFractions"][i]
        return out

    def CSb_Photo_CP(self, descriptor, energy):
        cp = self.CompoundParserCheckingNIST(descriptor,)
        out = 0.0
        for i in range(cp["nElements"]):
            out += self.CSb_Photo(cp["Elements"][i], energy) * cp["massFractions"][i]
        return out

    def CS_Rayl_CP(self, descriptor, energy):
        cp = self.CompoundParserCheckingNIST(descriptor,)
        out = 0.0
        for i in range(cp["nElements"]):
            out += self.CS_Rayl(cp["Elements"][i], energy) * cp["massFractions"][i]
        return out

    def CSb_Rayl_CP(self, descriptor, energy):
        cp = self.CompoundParserCheckingNIST(descriptor,)
        out = 0.0
        for i in range(cp["nElements"]):
            out += self.CSb_Rayl(cp["Elements"][i], energy) * cp["massFractions"][i]
        return out

    def CS_Compt_CP(self, descriptor, energy):
        cp = self.CompoundParserCheckingNIST(descriptor,)
        out = 0.0
        for i in range(cp["nElements"]):
            out += self.CS_Compt(cp["Elements"][i], energy) * cp["massFractions"][i]
        return out

    def CSb_Compt_CP(self, descriptor, energy):
        cp = self.CompoundParserCheckingNIST(descriptor,)
        out = 0.0
        for i in range(cp["nElements"]):
            out += self.CSb_Compt(cp["Elements"][i], energy) * cp["massFractions"][i]
        return out

    #########################
    #  refractive index
    #########################
    def Refractive_Index_Re(self, descriptor, energy, density):
        cp = self.compound_parser(descriptor)
        KD = 4.15179082788e-4  # TODO: recalculate with codata....
        rv = 0.0
        for i in range(cp["nElements"]):
            Z = cp["Elements"][i]
            rv += cp["massFractions"][i] * KD * (Z + self.Fi(Z, energy)) / \
                self.AtomicWeight(Z) / energy / energy
        return (1 - rv * density)

    def Refractive_Index_Im(self, descriptor, energy, density):
        cp = self.compound_parser(descriptor)
        rv = 0.0
        for i in range(cp["nElements"]):
            Z = cp["Elements"][i]
            rv += self.CS_Total(Z, energy) * cp["massFractions"][i]
        # /*9.8663479e-9 is calculated as planck's constant * speed of light / 4Pi */
        # return rv * density * 9.8663479e-9 / E;
        return rv * density * 9.8663479e-9 / energy

    #
    # NIST compounds
    #

    def GetCompoundDataNISTList(self):
        file1 = self.get_dabax_file("CompoundsNIST.dat")
        if self.verbose(): print("Accessing file: %s" % file1)
        sf = SpecFile(file1)

        list1 = []
        for index in range(len(sf)):
            list1.append(sf[index].scan_header_dict["Uname"])

        return list1

    def GetCompoundDataNISTByIndex(self, index_found):
        file1 = self.get_dabax_file("CompoundsNIST.dat")
        if self.verbose(): print("Accessing file: %s" % file1)
        sf = SpecFile(file1)

        s1 = sf[index_found]
        data = s1.data

        name = s1.scan_header_dict["Uname"]
        nElements = int(s1.scan_header_dict["UnElements"])
        density = float(s1.scan_header_dict["Udensity"])

        Elements = []
        massFractions = []
        for i in range(nElements):
            Elements.append(int(data[0][i]))
            massFractions.append(data[1][i])

        return {"name": name, 'nElements': nElements, 'density': density, 'Elements': Elements,
                'massFractions': massFractions}

    def GetCompoundDataNISTByName(self, entry_name):
        list1 = self.GetCompoundDataNISTList()
        return self.GetCompoundDataNISTByIndex(list1.index(entry_name))




    #
    #
    #   DONE:
    #
    #
    #  (used in xoppy_xraylib_util):
    #  xraylib.Crystal_GetCrystal(descriptor)
    #  xraylib.Crystal_dSpacing(cryst, hh, kk, ll)
    #  xraylib.Crystal_dSpacing
    #  xraylib.CompoundParser(descriptor)
    #  xraylib.SymbolToAtomicNumber(descriptor)
    #  xraylib.AtomicNumberToSymbol(zi)
    #  xraylib.ElementDensity(Z)
    #  xraylib.AtomicWeight
    #  xraylib.FF_Rayl(xraylib.SymbolToAtomicNumber(descriptor), iqscale)
    #  xraylib.Fi(Z,1e-3*ienergy)
    #  xraylib.Fii(Z,1e-3*ienergy)
    #  xraylib.Crystal_F_H_StructureFactor(_crystal, E_keV, h, k, l, _debyeWaller, 1.0)
    #  xraylib.Crystal_F_H_StructureFactor(_crystal, E_keV, h, k, l, _debyeWaller, 1.0)

    #
    #  (used in power/power3d)
    #
    #  xraylib.CS_Total(Z,1e-3*ienergy)
    #  xraylib.CSb_Total(Z,1e-3*ienergy)
    #  xraylib.CS_Total_CP(Z,1e-3*ienergy)
    #  xraylib.CSb_Total_CP(Z,1e-3*ienergy)
    #  xraylib.Refractive_Index_Re(descriptor, energy_in_keV, density)
    #  xraylib.Refractive_Index_Im(descriptor, energy_in_keV, density)
    #

    #  (used in calc_cross_sec)
    #  xraylib.CS_Phot()
    #  xraylib.CSb_Photo(Z,1e-3*ienergy)
    #  xraylib.CS_Phot_CP()
    #  xraylib.CSb_Photo_CP(descriptor,1e-3*ienergy)
    #  xraylib.CSb_Photo_CP(descriptor,1e-3*ienergy)
    #  xraylib.CS_Rayl(Z,1e-3*ienergy)
    #  xraylib.CSb_Rayl(Z,1e-3*ienergy)
    #  xraylib.CS_Rayl_CP(descriptor,1e-3*ienergy)
    #  xraylib.CSb_Rayl_CP(descriptor,1e-3*ienergy)
    #  xraylib.CS_Compt(Z,1e-3*ienergy)
    #  xraylib.CSb_Compt(Z,1e-3*ienergy)
    #  xraylib.CS_Compt_CP(descriptor,1e-3*ienergy)
    #  xraylib.CSb_Compt_CP(descriptor,1e-3*ienergy)
    #  xraylib.GetCompoundDataNISTList()
    #  xraylib.GetCompoundDataNISTByName(DESCRIPTOR)
    #  xraylib.GetCompoundDataNISTByIndex(DESCRIPTOR)

    #
    #   TODO
    #






    #
    # auxiliar methods
    # there are not in xraylib, but accelerate the calculation
    #
    def CompoundParserCheckingNIST(self, descriptor):
        try:
            out_dict = self.compound_parser(descriptor)

        except:
            try:
                out_dict = self.GetCompoundDataNISTByName(descriptor)
            except:
                raise Exception("Error processing compound descriptor: %s" % descriptor)

        return out_dict

    def FiAndFii(self, Z, energy):
        symbol = self.AtomicNumberToSymbol(Z)
        f1, f2 = self.f1f2_interpolate(symbol, energy*1e3)
        if self.get_file_f1f2() in ['f1f2_Windt.dat','f1f2_Henke.dat','f1f2_EPDL97.dat','f1f2_Chantler.dat','f1f2_asf_Kissel.dat']:
            f1 -= Z
        f2 *= -1.0
        return f1,f2

    def Crystal_F_0_F_H_F_H_bar_StructureFactor(self,
                                    crystal_id,
                                    energy_in_kev,
                                    millerH,
                                    millerK,
                                    millerL,
                                    debyeWaller,
                                    rel_angle=1.0):
        energy = energy_in_kev * 1e3
        wavelength = codata.h * codata.c / codata.e / energy * 1e10
        # print(crystal_id["n_atom"])
        atom = crystal_id['atom']
        natom = len(atom)
        list_fraction = [atom[i]['fraction'] for i in range(natom)]
        list_x = [atom[i]['x'] for i in range(natom)]
        list_y = [atom[i]['y'] for i in range(natom)]
        list_z = [atom[i]['z'] for i in range(natom)]

        F_0 = numpy.zeros(numpy.array(energy).size, dtype=complex)
        F_H = numpy.zeros(numpy.array(energy).size, dtype=complex)
        F_H_bar = numpy.zeros(numpy.array(energy).size, dtype=complex)

        for i in range(natom):
            atom_i = atom[i]
            if (i > 0) and (atom_i['Zatom'] == Z_i) and (atom_i['charge'] == charge_i):
                pass # avoid re-calculating f0 if inputs are identical to previous call
            else:
                Z_i = atom_i['Zatom']
                charge_i = atom_i['charge']
                coeffs = self.f0_with_fractional_charge(Z_i, charge=charge_i)


                if (millerH == 0 and millerK == 0 and millerL == 0):
                    ratio = 0.0
                else:
                    angle = self.Bragg_angle(crystal_id, energy_in_kev,
                                             millerH, millerK, millerL)
                    ratio = numpy.sin(angle * rel_angle) / wavelength

                f0_i_zero = calculate_f0_from_f0coeff(coeffs, 0.0)
                f0_i = calculate_f0_from_f0coeff(coeffs, ratio)
                Fi  = self.Fi(Z_i, energy_in_kev)
                Fii = self.Fii(Z_i, energy_in_kev)

            F_0 += (f0_i_zero + Fi - Fii * 1j) * list_fraction[i] * debyeWaller

            F_H += (f0_i + Fi - Fii * 1j) * list_fraction[i] * debyeWaller * \
                   numpy.exp(2j*numpy.pi*(millerH*list_x[i]+millerK*list_y[i]+millerL*list_z[i]))

            F_H_bar += (f0_i + Fi - Fii * 1j) * list_fraction[i] * debyeWaller * \
                   numpy.exp(2j*numpy.pi*(-millerH*list_x[i]-millerK*list_y[i]-millerL*list_z[i]))

        return F_0, F_H, F_H_bar






# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/


__authors__ = ["M Sanchez del Rio - ESRF"]
__license__ = "MIT"
__date__ = "22/10/2021"

from setuptools import setup, find_packages

setup(name='dabax',
    version='1.0.7',
    description='python access to DABAX files',
    author='M Sanchez del Rio',
    author_email='srio@esrf.eu',
    url='https://github.com/oasys-kit/dabax/',
    packages=find_packages(include=['dabax', 'dabax.*']),
    install_requires=[
        'numpy',
        'scipy',
        'silx',
        ],
    setup_requires=[
        'setuptools',
        ],
    )
